# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#                      Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import os
import shutil
import tempfile
from babeltrace import CTFWriter, CTFStringEncoding


class TraceWriter():
    def __init__(self):
        self._trace_root = tempfile.mkdtemp()
        self.trace_path = os.path.join(self.trace_root, "kernel")
        self.create_writer()
        self.create_stream_class()
        self.define_base_types()
        self.define_events()
        self.create_stream()

    @property
    def trace_root(self):
        return self._trace_root

    def rm_trace(self):
        shutil.rmtree(self.trace_root)

    def flush(self):
        self.writer.flush_metadata()
        self.stream.flush()

    def create_writer(self):
        self.clock = CTFWriter.Clock("A_clock")
        self.clock.description = "Simple clock"
        self.writer = CTFWriter.Writer(self.trace_path)
        self.writer.add_clock(self.clock)
        self.writer.add_environment_field("Python_version",
                                          str(sys.version_info))
        self.writer.add_environment_field("tracer_major", 2)
        self.writer.add_environment_field("tracer_minor", 8)
        self.writer.add_environment_field("tracer_patchlevel", 0)

    def create_stream_class(self):
        self.stream_class = CTFWriter.StreamClass("test_stream")
        self.stream_class.clock = self.clock

    def define_base_types(self):
        self.char8_type = CTFWriter.IntegerFieldDeclaration(8)
        self.char8_type.signed = True
        self.char8_type.encoding = CTFStringEncoding.UTF8
        self.char8_type.alignment = 8

        self.int16_type = CTFWriter.IntegerFieldDeclaration(16)
        self.int16_type.signed = True
        self.int16_type.alignment = 8

        self.uint16_type = CTFWriter.IntegerFieldDeclaration(16)
        self.uint16_type.signed = False
        self.uint16_type.alignment = 8

        self.int32_type = CTFWriter.IntegerFieldDeclaration(32)
        self.int32_type.signed = True
        self.int32_type.alignment = 8

        self.uint32_type = CTFWriter.IntegerFieldDeclaration(32)
        self.uint32_type.signed = False
        self.uint32_type.alignment = 8

        self.int64_type = CTFWriter.IntegerFieldDeclaration(64)
        self.int64_type.signed = True
        self.int64_type.alignment = 8

        self.uint64_type = CTFWriter.IntegerFieldDeclaration(64)
        self.uint64_type.signed = False
        self.uint64_type.alignment = 8

        self.array16_type = CTFWriter.ArrayFieldDeclaration(self.char8_type,
                                                            16)

        self.string_type = CTFWriter.StringFieldDeclaration()

    def add_event(self, event):
        event.add_field(self.uint32_type, "_cpu_id")
        self.stream_class.add_event_class(event)

    def define_sched_switch(self):
        self.sched_switch = CTFWriter.EventClass("sched_switch")
        self.sched_switch.add_field(self.array16_type, "_prev_comm")
        self.sched_switch.add_field(self.int32_type, "_prev_tid")
        self.sched_switch.add_field(self.int32_type, "_prev_prio")
        self.sched_switch.add_field(self.int64_type, "_prev_state")
        self.sched_switch.add_field(self.array16_type, "_next_comm")
        self.sched_switch.add_field(self.int32_type, "_next_tid")
        self.sched_switch.add_field(self.int32_type, "_next_prio")
        self.add_event(self.sched_switch)

    def define_softirq_raise(self):
        self.softirq_raise = CTFWriter.EventClass("softirq_raise")
        self.softirq_raise.add_field(self.uint32_type, "_vec")
        self.add_event(self.softirq_raise)

    def define_softirq_entry(self):
        self.softirq_entry = CTFWriter.EventClass("softirq_entry")
        self.softirq_entry.add_field(self.uint32_type, "_vec")
        self.add_event(self.softirq_entry)

    def define_softirq_exit(self):
        self.softirq_exit = CTFWriter.EventClass("softirq_exit")
        self.softirq_exit.add_field(self.uint32_type, "_vec")
        self.add_event(self.softirq_exit)

    def define_irq_handler_entry(self):
        self.irq_handler_entry = CTFWriter.EventClass("irq_handler_entry")
        self.irq_handler_entry.add_field(self.int32_type, "_irq")
        self.irq_handler_entry.add_field(self.string_type, "_name")
        self.add_event(self.irq_handler_entry)

    def define_irq_handler_exit(self):
        self.irq_handler_exit = CTFWriter.EventClass("irq_handler_exit")
        self.irq_handler_exit.add_field(self.int32_type, "_irq")
        self.irq_handler_exit.add_field(self.int32_type, "_ret")
        self.add_event(self.irq_handler_exit)

    def define_syscall_entry_write(self):
        self.syscall_entry_write = CTFWriter.EventClass("syscall_entry_write")
        self.syscall_entry_write.add_field(self.uint32_type, "_fd")
        self.syscall_entry_write.add_field(self.uint64_type, "_buf")
        self.syscall_entry_write.add_field(self.uint64_type, "_count")
        self.add_event(self.syscall_entry_write)

    def define_syscall_exit_write(self):
        self.syscall_exit_write = CTFWriter.EventClass("syscall_exit_write")
        self.syscall_exit_write.add_field(self.int64_type, "_ret")
        self.add_event(self.syscall_exit_write)

    def define_syscall_entry_read(self):
        self.syscall_entry_read = CTFWriter.EventClass("syscall_entry_read")
        self.syscall_entry_read.add_field(self.uint32_type, "_fd")
        self.syscall_entry_read.add_field(self.uint64_type, "_count")
        self.add_event(self.syscall_entry_read)

    def define_syscall_exit_read(self):
        self.syscall_exit_read = CTFWriter.EventClass("syscall_exit_read")
        self.syscall_exit_read.add_field(self.uint64_type, "_buf")
        self.syscall_exit_read.add_field(self.int64_type, "_ret")
        self.add_event(self.syscall_exit_read)

    def define_syscall_entry_open(self):
        self.syscall_entry_open = CTFWriter.EventClass("syscall_entry_open")
        self.syscall_entry_open.add_field(self.string_type, "_filename")
        self.syscall_entry_open.add_field(self.int32_type, "_flags")
        self.syscall_entry_open.add_field(self.uint16_type, "_mode")
        self.add_event(self.syscall_entry_open)

    def define_syscall_exit_open(self):
        self.syscall_exit_open = CTFWriter.EventClass("syscall_exit_open")
        self.syscall_exit_open.add_field(self.int64_type, "_ret")
        self.add_event(self.syscall_exit_open)

    def define_lttng_statedump_process_state(self):
        self.lttng_statedump_process_state = CTFWriter.EventClass(
            "lttng_statedump_process_state")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_tid")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_vtid")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_pid")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_vpid")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_ppid")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_vppid")
        self.lttng_statedump_process_state.add_field(self.array16_type,
                                                     "_name")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_type")
        self.lttng_statedump_process_state.add_field(self.int32_type, "_mode")
        self.lttng_statedump_process_state.add_field(self.int32_type,
                                                     "_submode")
        self.lttng_statedump_process_state.add_field(self.int32_type,
                                                     "_status")
        self.lttng_statedump_process_state.add_field(self.int32_type,
                                                     "_ns_level")
        self.add_event(self.lttng_statedump_process_state)

    def define_lttng_statedump_file_descriptor(self):
        self.lttng_statedump_file_descriptor = CTFWriter.EventClass(
            "lttng_statedump_file_descriptor")
        self.lttng_statedump_file_descriptor.add_field(self.int32_type, "_pid")
        self.lttng_statedump_file_descriptor.add_field(self.int32_type, "_fd")
        self.lttng_statedump_file_descriptor.add_field(self.uint32_type,
                                                       "_flags")
        self.lttng_statedump_file_descriptor.add_field(self.uint32_type,
                                                       "_fmode")
        self.lttng_statedump_file_descriptor.add_field(self.string_type,
                                                       "_filename")
        self.add_event(self.lttng_statedump_file_descriptor)

    def define_sched_wakeup(self):
        self.sched_wakeup = CTFWriter.EventClass("sched_wakeup")
        self.sched_wakeup.add_field(self.array16_type, "_comm")
        self.sched_wakeup.add_field(self.int32_type, "_tid")
        self.sched_wakeup.add_field(self.int32_type, "_prio")
        self.sched_wakeup.add_field(self.int32_type, "_success")
        self.sched_wakeup.add_field(self.int32_type, "_target_cpu")
        self.add_event(self.sched_wakeup)

    def define_sched_waking(self):
        self.sched_waking = CTFWriter.EventClass("sched_waking")
        self.sched_waking.add_field(self.array16_type, "_comm")
        self.sched_waking.add_field(self.int32_type, "_tid")
        self.sched_waking.add_field(self.int32_type, "_prio")
        self.sched_waking.add_field(self.int32_type, "_target_cpu")
        self.add_event(self.sched_waking)

    def define_block_rq_complete(self):
        self.block_rq_complete = CTFWriter.EventClass("block_rq_complete")
        self.block_rq_complete.add_field(self.uint32_type, "_dev")
        self.block_rq_complete.add_field(self.uint64_type, "_sector")
        self.block_rq_complete.add_field(self.uint32_type, "_nr_sector")
        self.block_rq_complete.add_field(self.int32_type, "_errors")
        self.block_rq_complete.add_field(self.uint32_type, "_rwbs")
        self.block_rq_complete.add_field(self.uint64_type, "__cmd_length")
        self.block_rq_complete.add_field(self.array16_type, "_cmd")
        self.add_event(self.block_rq_complete)

    def define_block_rq_issue(self):
        self.block_rq_issue = CTFWriter.EventClass("block_rq_issue")
        self.block_rq_issue.add_field(self.uint32_type, "_dev")
        self.block_rq_issue.add_field(self.uint64_type, "_sector")
        self.block_rq_issue.add_field(self.uint32_type, "_nr_sector")
        self.block_rq_issue.add_field(self.uint32_type, "_bytes")
        self.block_rq_issue.add_field(self.int32_type, "_tid")
        self.block_rq_issue.add_field(self.uint32_type, "_rwbs")
        self.block_rq_issue.add_field(self.uint64_type, "__cmd_length")
        self.block_rq_issue.add_field(self.array16_type, "_cmd")
        self.block_rq_issue.add_field(self.array16_type, "_comm")
        self.add_event(self.block_rq_issue)

    def define_net_dev_xmit(self):
        self.net_dev_xmit = CTFWriter.EventClass("net_dev_xmit")
        self.net_dev_xmit.add_field(self.uint64_type, "_skbaddr")
        self.net_dev_xmit.add_field(self.int32_type, "_rc")
        self.net_dev_xmit.add_field(self.uint32_type, "_len")
        self.net_dev_xmit.add_field(self.string_type, "_name")
        self.add_event(self.net_dev_xmit)

    def define_netif_receive_skb(self):
        self.netif_receive_skb = CTFWriter.EventClass("netif_receive_skb")
        self.netif_receive_skb.add_field(self.uint64_type, "_skbaddr")
        self.netif_receive_skb.add_field(self.uint32_type, "_len")
        self.netif_receive_skb.add_field(self.string_type, "_name")
        self.add_event(self.netif_receive_skb)

    def define_events(self):
        self.define_sched_switch()
        self.define_softirq_raise()
        self.define_softirq_entry()
        self.define_softirq_exit()
        self.define_irq_handler_entry()
        self.define_irq_handler_exit()
        self.define_syscall_entry_write()
        self.define_syscall_exit_write()
        self.define_syscall_entry_read()
        self.define_syscall_exit_read()
        self.define_syscall_entry_open()
        self.define_syscall_exit_open()
        self.define_lttng_statedump_process_state()
        self.define_lttng_statedump_file_descriptor()
        self.define_sched_wakeup()
        self.define_sched_waking()
        self.define_block_rq_complete()
        self.define_block_rq_issue()
        self.define_net_dev_xmit()
        self.define_netif_receive_skb()

    def create_stream(self):
        self.stream = self.writer.create_stream(self.stream_class)

    def set_char_array(self, event, string):
        if len(string) > 16:
            string = string[0:16]
        else:
            string = "%s" % (string + "\0" * (16 - len(string)))

        for i, char in enumerate(string):
            event.field(i).value = ord(char)

    def set_int(self, event, value):
        event.value = value

    def set_string(self, event, value):
        event.value = value

    def write_softirq_raise(self, time_ms, cpu_id, vec):
        event = CTFWriter.Event(self.softirq_raise)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_vec"), vec)
        self.stream.append_event(event)
        self.stream.flush()

    def write_softirq_entry(self, time_ms, cpu_id, vec):
        event = CTFWriter.Event(self.softirq_entry)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_vec"), vec)
        self.stream.append_event(event)
        self.stream.flush()

    def write_softirq_exit(self, time_ms, cpu_id, vec):
        event = CTFWriter.Event(self.softirq_exit)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_vec"), vec)
        self.stream.append_event(event)
        self.stream.flush()

    def write_irq_handler_entry(self, time_ms, cpu_id, irq, name):
        event = CTFWriter.Event(self.irq_handler_entry)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_irq"), irq)
        self.set_string(event.payload("_name"), name)
        self.stream.append_event(event)
        self.stream.flush()

    def write_irq_handler_exit(self, time_ms, cpu_id, irq, ret):
        event = CTFWriter.Event(self.irq_handler_exit)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_irq"), irq)
        self.set_int(event.payload("_ret"), ret)
        self.stream.append_event(event)
        self.stream.flush()

    def write_syscall_write(self, time_ms, cpu_id, delay, fd, buf, count, ret):
        event_entry = CTFWriter.Event(self.syscall_entry_write)
        self.clock.time = time_ms * 1000000
        self.set_int(event_entry.payload("_cpu_id"), cpu_id)
        self.set_int(event_entry.payload("_fd"), fd)
        self.set_int(event_entry.payload("_buf"), buf)
        self.set_int(event_entry.payload("_count"), count)
        self.stream.append_event(event_entry)

        event_exit = CTFWriter.Event(self.syscall_exit_write)
        self.clock.time = (time_ms + delay) * 1000000
        self.set_int(event_exit.payload("_cpu_id"), cpu_id)
        self.set_int(event_exit.payload("_ret"), ret)
        self.stream.append_event(event_exit)
        self.stream.flush()

    def write_syscall_read(self, time_ms, cpu_id, delay, fd, buf, count, ret):
        event_entry = CTFWriter.Event(self.syscall_entry_read)
        self.clock.time = time_ms * 1000000
        self.set_int(event_entry.payload("_cpu_id"), cpu_id)
        self.set_int(event_entry.payload("_fd"), fd)
        self.set_int(event_entry.payload("_count"), count)
        self.stream.append_event(event_entry)

        event_exit = CTFWriter.Event(self.syscall_exit_read)
        self.clock.time = (time_ms + delay) * 1000000
        self.set_int(event_exit.payload("_cpu_id"), cpu_id)
        self.set_int(event_exit.payload("_buf"), buf)
        self.set_int(event_exit.payload("_ret"), ret)
        self.stream.append_event(event_exit)
        self.stream.flush()

    def write_syscall_open(self, time_ms, cpu_id, delay, filename, flags,
                           mode, ret):
        event = CTFWriter.Event(self.syscall_entry_open)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_string(event.payload("_filename"), filename)
        self.set_int(event.payload("_flags"), flags)
        self.set_int(event.payload("_mode"), mode)
        self.stream.append_event(event)
        self.stream.flush()

        event = CTFWriter.Event(self.syscall_exit_open)
        self.clock.time = (time_ms + delay) * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_ret"), ret)
        self.stream.append_event(event)
        self.stream.flush()

    def write_lttng_statedump_file_descriptor(self, time_ms, cpu_id, pid, fd,
                                              flags, fmode, filename):
        event = CTFWriter.Event(self.lttng_statedump_file_descriptor)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_pid"), pid)
        self.set_int(event.payload("_fd"), fd)
        self.set_int(event.payload("_flags"), flags)
        self.set_int(event.payload("_fmode"), fmode)
        self.set_string(event.payload("_filename"), filename)
        self.stream.append_event(event)
        self.stream.flush()

    def write_lttng_statedump_process_state(self, time_ms, cpu_id, tid, vtid,
                                            pid, vpid, ppid, vppid, name, type,
                                            mode, submode, status, ns_level):
        event = CTFWriter.Event(self.lttng_statedump_process_state)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_tid"), tid)
        self.set_int(event.payload("_vtid"), vtid)
        self.set_int(event.payload("_pid"), pid)
        self.set_int(event.payload("_vpid"), vpid)
        self.set_int(event.payload("_ppid"), ppid)
        self.set_int(event.payload("_vppid"), vppid)
        self.set_char_array(event.payload("_name"), name)
        self.set_int(event.payload("_type"), type)
        self.set_int(event.payload("_mode"), mode)
        self.set_int(event.payload("_submode"), submode)
        self.set_int(event.payload("_status"), status)
        self.set_int(event.payload("_ns_level"), ns_level)
        self.stream.append_event(event)
        self.stream.flush()

    def write_sched_wakeup(self, time_ms, cpu_id, comm, tid, prio, target_cpu):
        event = CTFWriter.Event(self.sched_wakeup)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_char_array(event.payload("_comm"), comm)
        self.set_int(event.payload("_tid"), tid)
        self.set_int(event.payload("_prio"), prio)
        self.set_int(event.payload("_target_cpu"), target_cpu)
        self.stream.append_event(event)
        self.stream.flush()

    def write_sched_waking(self, time_ms, cpu_id, comm, tid, prio, target_cpu):
        event = CTFWriter.Event(self.sched_waking)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_char_array(event.payload("_comm"), comm)
        self.set_int(event.payload("_tid"), tid)
        self.set_int(event.payload("_prio"), prio)
        self.set_int(event.payload("_target_cpu"), target_cpu)
        self.stream.append_event(event)
        self.stream.flush()

    def write_block_rq_complete(self, time_ms, cpu_id, dev, sector, nr_sector,
                                errors, rwbs, _cmd_length, cmd):
        event = CTFWriter.Event(self.block_rq_complete)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_dev"), dev)
        self.set_int(event.payload("_sector"), sector)
        self.set_int(event.payload("_nr_sector"), nr_sector)
        self.set_int(event.payload("_errors"), errors)
        self.set_int(event.payload("_rwbs"), rwbs)
        self.set_int(event.payload("__cmd_length"), _cmd_length)
        self.set_char_array(event.payload("_cmd"), cmd)
        self.stream.append_event(event)
        self.stream.flush()

    def write_block_rq_issue(self, time_ms, cpu_id, dev, sector, nr_sector,
                             bytes, tid, rwbs, _cmd_length, cmd, comm):
        event = CTFWriter.Event(self.block_rq_issue)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_dev"), dev)
        self.set_int(event.payload("_sector"), sector)
        self.set_int(event.payload("_nr_sector"), nr_sector)
        self.set_int(event.payload("_bytes"), bytes)
        self.set_int(event.payload("_tid"), tid)
        self.set_int(event.payload("_rwbs"), rwbs)
        self.set_int(event.payload("__cmd_length"), _cmd_length)
        self.set_char_array(event.payload("_cmd"), cmd)
        self.set_char_array(event.payload("_comm"), comm)
        self.stream.append_event(event)
        self.stream.flush()

    def write_net_dev_xmit(self, time_ms, cpu_id, skbaddr, rc, len, name):
        event = CTFWriter.Event(self.net_dev_xmit)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_skbaddr"), skbaddr)
        self.set_int(event.payload("_rc"), rc)
        self.set_int(event.payload("_len"), len)
        self.set_string(event.payload("_name"), name)
        self.stream.append_event(event)
        self.stream.flush()

    def write_netif_receive_skb(self, time_ms, cpu_id, skbaddr, len, name):
        event = CTFWriter.Event(self.netif_receive_skb)
        self.clock.time = time_ms * 1000000
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.set_int(event.payload("_skbaddr"), skbaddr)
        self.set_int(event.payload("_len"), len)
        self.set_string(event.payload("_name"), name)
        self.stream.append_event(event)
        self.stream.flush()

    def write_sched_switch(self, time_ms, cpu_id, prev_comm, prev_tid,
                           next_comm, next_tid, prev_prio=20, prev_state=1,
                           next_prio=20):
        event = CTFWriter.Event(self.sched_switch)
        self.clock.time = time_ms * 1000000
        self.set_char_array(event.payload("_prev_comm"), prev_comm)
        self.set_int(event.payload("_prev_tid"), prev_tid)
        self.set_int(event.payload("_prev_prio"), prev_prio)
        self.set_int(event.payload("_prev_state"), prev_state)
        self.set_char_array(event.payload("_next_comm"), next_comm)
        self.set_int(event.payload("_next_tid"), next_tid)
        self.set_int(event.payload("_next_prio"), next_prio)
        self.set_int(event.payload("_cpu_id"), cpu_id)
        self.stream.append_event(event)
        self.stream.flush()

    def sched_switch_50pc(self, start_time_ms, end_time_ms, cpu_id, period,
                          comm1, tid1, comm2, tid2):
        current = start_time_ms
        while current < end_time_ms:
            self.write_sched_switch(current, cpu_id, comm1, tid1, comm2, tid2)
            current += period
            self.write_sched_switch(current, cpu_id, comm2, tid2, comm1, tid1)
            current += period



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import time


class TimezoneUtils():
    def __init__(self):
        self.original_tz = None

    def set_up_timezone(self):
        # Make sure that the local timezone as seen by the time module
        # is the same regardless of where the test is actually
        # run. US/Eastern was picked arbitrarily.
        self.original_tz = os.environ.get('TZ')
        os.environ['TZ'] = 'US/Eastern'
        try:
            time.tzset()
        except AttributeError:
            print('Warning: time.tzset() is unavailable on Windows.'
                  'This may cause test failures.')

    def tear_down_timezone(self):
        # Restore the original value of TZ if any, else delete it from
        # the environment variables.
        if self.original_tz:
            os.environ['TZ'] = self.original_tz
        else:
            del os.environ['TZ']



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import datetime
import unittest
from lttnganalyses.common import parse_utils
from .utils import TimezoneUtils


# Mock of babeltrace's TraceCollection, used to test date methods
class TraceCollection():
    def __init__(self, begin_ts, end_ts):
        self.begin_ts = begin_ts
        self.end_ts = end_ts

    @property
    def timestamp_begin(self):
        return self.begin_ts

    @property
    def timestamp_end(self):
        return self.end_ts


class TestParseSize(unittest.TestCase):
    def test_garbage(self):
        self.assertRaises(ValueError, parse_utils.parse_size,
                          'ceci n\'est pas une size')
        self.assertRaises(ValueError, parse_utils.parse_size,
                          '12.34.56')

    def test_invalid_units(self):
        self.assertRaises(ValueError, parse_utils.parse_size, '500 furlongs')

    def test_binary_units(self):
        result = parse_utils.parse_size('500 KiB')
        self.assertEqual(result, 512000)
        result = parse_utils.parse_size('-500 KiB')
        self.assertEqual(result, -512000)
        # no space left between units and value is intentional
        result = parse_utils.parse_size('0.01MiB')
        self.assertEqual(result, 10485)
        result = parse_utils.parse_size('1200 YiB')
        self.assertEqual(result, 1450710983537555009647411200)
        result = parse_utils.parse_size('1234 B')
        self.assertEqual(result, 1234)

    def test_coreutils_units(self):
        result = parse_utils.parse_size('500 K')
        self.assertEqual(result, 512000)
        result = parse_utils.parse_size('-500 K')
        self.assertEqual(result, -512000)
        # no space left between units and value is intentional
        result = parse_utils.parse_size('0.01M')
        self.assertEqual(result, 10485)
        result = parse_utils.parse_size('1200 Y')
        self.assertEqual(result, 1450710983537555009647411200)

    def test_si_units(self):
        result = parse_utils.parse_size('500 KB')
        self.assertEqual(result, 500000)
        result = parse_utils.parse_size('-500 KB')
        self.assertEqual(result, -500000)
        # no space left between units and value is intentional
        result = parse_utils.parse_size('0.01MB')
        self.assertEqual(result, 10000)
        result = parse_utils.parse_size('40 ZB')
        self.assertEqual(result, 40000000000000000000000)
        # Sizes a bit larger than 40 ZB (e.g. 50 ZB and up) with
        # decimal units don't get parsed quite as precisely because of
        # the nature of floating point numbers. If precision is needed
        # for larger values with these units, it could be fixed, but
        # for now it seems unlikely so we leave it as is

    def test_no_units(self):
        result = parse_utils.parse_size('1234')
        self.assertEqual(result, 1234)
        result = parse_utils.parse_size('1234.567')
        self.assertEqual(result, 1234)
        result = parse_utils.parse_size('-1234.567')
        self.assertEqual(result, -1234)


class TestParseDuration(unittest.TestCase):
    def test_garbage(self):
        self.assertRaises(ValueError, parse_utils.parse_duration,
                          'ceci n\'est pas une duration')
        self.assertRaises(ValueError, parse_utils.parse_duration,
                          '12.34.56')

    def test_invalid_units(self):
        self.assertRaises(ValueError, parse_utils.parse_duration,
                          '500 furlongs')

    def test_valid_units(self):
        result = parse_utils.parse_duration('1s')
        self.assertEqual(result, 1000000000)
        result = parse_utils.parse_duration('-1s')
        self.assertEqual(result, -1000000000)
        result = parse_utils.parse_duration('1234.56 ms')
        self.assertEqual(result, 1234560000)
        result = parse_utils.parse_duration('1.23 us')
        self.assertEqual(result, 1230)
        result = parse_utils.parse_duration('1.23 µs')
        self.assertEqual(result, 1230)
        result = parse_utils.parse_duration('1234 ns')
        self.assertEqual(result, 1234)
        result = parse_utils.parse_duration('0.001 ns')
        self.assertEqual(result, 0)

    def test_no_units(self):
        result = parse_utils.parse_duration('1234.567')
        self.assertEqual(result, 1234567000000)


class TestParseDate(unittest.TestCase):
    def setUp(self):
        self.tz_utils = TimezoneUtils()
        self.tz_utils.set_up_timezone()

    def tearDown(self):
        self.tz_utils.tear_down_timezone()

    def test_parse_full_date_nsec(self):
        date_expected = datetime.datetime(2014, 12, 12, 17, 29, 43)
        nsec_expected = 802588035
        date, nsec = parse_utils.parse_date('2014-12-12 17:29:43.802588035')
        self.assertEqual(date, date_expected)
        self.assertEqual(nsec, nsec_expected)
        date, nsec = parse_utils.parse_date('2014-12-12T17:29:43.802588035')
        self.assertEqual(date, date_expected)
        self.assertEqual(nsec, nsec_expected)

    def test_parse_full_date(self):
        date_expected = datetime.datetime(2014, 12, 12, 17, 29, 43)
        nsec_expected = 0
        date, nsec = parse_utils.parse_date('2014-12-12 17:29:43')
        self.assertEqual(date, date_expected)
        self.assertEqual(nsec, nsec_expected)
        date, nsec = parse_utils.parse_date('2014-12-12T17:29:43')
        self.assertEqual(date, date_expected)
        self.assertEqual(nsec, nsec_expected)

    def test_parse_time_nsec(self):
        time_expected = datetime.time(17, 29, 43)
        nsec_expected = 802588035
        time, nsec = parse_utils.parse_date('17:29:43.802588035')
        self.assertEqual(time, time_expected)
        self.assertEqual(nsec, nsec_expected)

    def test_parse_time(self):
        time_expected = datetime.time(17, 29, 43)
        nsec_expected = 0
        time, nsec = parse_utils.parse_date('17:29:43')
        self.assertEqual(time, time_expected)
        self.assertEqual(nsec, nsec_expected)

    def test_parse_timestamp(self):
        time_expected = datetime.datetime(2014, 12, 12, 17, 29, 43)
        nsec_expected = 802588035
        date, nsec = parse_utils.parse_date('1418423383802588035')
        self.assertEqual(date, time_expected)
        self.assertEqual(nsec, nsec_expected)

    def test_parse_date_invalid(self):
        self.assertRaises(ValueError, parse_utils.parse_date,
                          'ceci n\'est pas une date')


class TestParseTraceCollectionDate(unittest.TestCase):
    DATE_FULL = '2014-12-12 17:29:43'
    DATE_TIME = '17:29:43'
    SINGLE_DAY_COLLECTION = TraceCollection(
        1418423383802588035, 1418423483802588035
    )
    MULTI_DAY_COLLECTION = TraceCollection(
        1418423383802588035, 1419423383802588035
    )

    def _mock_parse_date(self, date):
        if date == self.DATE_FULL:
            return (datetime.datetime(2014, 12, 12, 17, 29, 43), 0)
        elif date == self.DATE_TIME:
            return (datetime.time(17, 29, 43), 0)
        else:
            raise ValueError('Unrecognised date format: {}'.format(date))

    def setUp(self):
        self.tz_utils = TimezoneUtils()
        self.tz_utils.set_up_timezone()
        self._original_parse_date = parse_utils.parse_date
        parse_utils.parse_date = self._mock_parse_date

    def tearDown(self):
        self.tz_utils.tear_down_timezone()
        parse_utils.parse_date = self._original_parse_date

    def test_invalid_date(self):
        self.assertRaises(
            ValueError, parse_utils.parse_trace_collection_date,
            self.SINGLE_DAY_COLLECTION, 'ceci n\'est pas une date'
        )

    def test_single_day_date(self):
        expected = 1418423383000000000
        result = parse_utils.parse_trace_collection_date(
            self.SINGLE_DAY_COLLECTION, self.DATE_FULL
        )
        self.assertEqual(result, expected)

    def test_single_day_time(self):
        expected = 1418423383000000000
        result = parse_utils.parse_trace_collection_date(
            self.SINGLE_DAY_COLLECTION, self.DATE_TIME
        )
        self.assertEqual(result, expected)

    def test_multi_day_date(self):
        expected = 1418423383000000000
        result = parse_utils.parse_trace_collection_date(
            self.MULTI_DAY_COLLECTION, self.DATE_FULL
        )
        self.assertEqual(result, expected)

    def test_multi_day_time(self):
        self.assertRaises(
            ValueError, parse_utils.parse_trace_collection_date,
            self.MULTI_DAY_COLLECTION, self.DATE_TIME
        )


class TestParseTraceCollectionTimeRange(unittest.TestCase):
    DATE_FULL_BEGIN = '2014-12-12 17:29:43'
    DATE_FULL_END = '2014-12-12 17:29:44'
    DATE_TIME_BEGIN = '17:29:43'
    DATE_TIME_END = '17:29:44'
    EXPECTED_BEGIN = 1418423383000000000
    EXPECTED_END = 1418423384000000000
    SINGLE_DAY_COLLECTION = TraceCollection(
        1418423383802588035, 1418423483802588035
    )
    MULTI_DAY_COLLECTION = TraceCollection(
        1418423383802588035, 1419423383802588035
    )
    TIME_RANGE_FMT = '[{}, {}]'

    def _mock_parse_trace_collection_date(self, collection, date, gmt=False,
                                          handles=None):
        if collection == self.SINGLE_DAY_COLLECTION:
            if date == self.DATE_FULL_BEGIN or date == self.DATE_TIME_BEGIN:
                timestamp = 1418423383000000000
            elif date == self.DATE_FULL_END or date == self.DATE_TIME_END:
                timestamp = 1418423384000000000
            else:
                raise ValueError('Unrecognised date format: {}'.format(date))
        elif collection == self.MULTI_DAY_COLLECTION:
            if date == self.DATE_FULL_BEGIN:
                timestamp = 1418423383000000000
            elif date == self.DATE_FULL_END:
                timestamp = 1418423384000000000
            elif date == self.DATE_TIME_BEGIN or date == self.DATE_TIME_END:
                raise ValueError(
                    'Invalid date format for multi-day trace: {}'.format(date)
                )
            else:
                raise ValueError('Unrecognised date format: {}'.format(date))

        return timestamp

    def setUp(self):
        self._original_parse_trace_collection_date = (
            parse_utils.parse_trace_collection_date
        )
        parse_utils.parse_trace_collection_date = (
            self._mock_parse_trace_collection_date
        )

    def tearDown(self):
        parse_utils.parse_trace_collection_date = (
            self._original_parse_trace_collection_date
        )

    def test_invalid_format(self):
        self.assertRaises(
            ValueError, parse_utils.parse_trace_collection_time_range,
            self.SINGLE_DAY_COLLECTION, 'ceci n\'est pas un time range'
        )

    def test_single_day_date(self):
        time_range = self.TIME_RANGE_FMT.format(
            self.DATE_FULL_BEGIN, self.DATE_FULL_END
        )
        begin, end = parse_utils.parse_trace_collection_time_range(
            self.SINGLE_DAY_COLLECTION, time_range
        )
        self.assertEqual(begin, self.EXPECTED_BEGIN)
        self.assertEqual(end, self.EXPECTED_END)

    def test_single_day_time(self):
        time_range = self.TIME_RANGE_FMT.format(
            self.DATE_TIME_BEGIN, self.DATE_TIME_END
        )
        begin, end = parse_utils.parse_trace_collection_time_range(
            self.SINGLE_DAY_COLLECTION, time_range
        )
        self.assertEqual(begin, self.EXPECTED_BEGIN)
        self.assertEqual(end, self.EXPECTED_END)

    def test_multi_day_date(self):
        time_range = self.TIME_RANGE_FMT.format(
            self.DATE_FULL_BEGIN, self.DATE_FULL_END
        )
        begin, end = parse_utils.parse_trace_collection_time_range(
            self.MULTI_DAY_COLLECTION, time_range
        )
        self.assertEqual(begin, self.EXPECTED_BEGIN)
        self.assertEqual(end, self.EXPECTED_END)

    def test_multi_day_time(self):
        time_range = self.TIME_RANGE_FMT.format(
            self.DATE_TIME_BEGIN, self.DATE_TIME_END
        )
        self.assertRaises(
            ValueError, parse_utils.parse_trace_collection_time_range,
            self.MULTI_DAY_COLLECTION, time_range
        )



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import sp, sv
from ..common import version_utils


class SchedStateProvider(sp.StateProvider):
    # The priority offset for sched_wak* events was fixed in
    # lttng-modules 2.7.1 upwards
    PRIO_OFFSET_FIX_VERSION = version_utils.Version(2, 7, 1)

    def __init__(self, state):
        cbs = {
            'sched_switch': self._process_sched_switch,
            'sched_migrate_task': self._process_sched_migrate_task,
            'sched_wakeup': self._process_sched_wakeup,
            'sched_wakeup_new': self._process_sched_wakeup,
            'sched_waking': self._process_sched_wakeup,
            'sched_process_fork': self._process_sched_process_fork,
            'sched_process_exec': self._process_sched_process_exec,
            'sched_pi_setprio': self._process_sched_pi_setprio,
        }

        super().__init__(state, cbs)

    def _sched_switch_per_cpu(self, cpu_id, next_tid):
        if cpu_id not in self._state.cpus:
            self._state.cpus[cpu_id] = sv.CPU(cpu_id)

        cpu = self._state.cpus[cpu_id]
        # exclude swapper process
        if next_tid == 0:
            cpu.current_tid = None
        else:
            cpu.current_tid = next_tid

    def _create_proc(self, tid):
        if tid not in self._state.tids:
            if tid == 0:
                # special case for the swapper
                self._state.tids[tid] = sv.Process(tid=tid, pid=0)
            else:
                self._state.tids[tid] = sv.Process(tid=tid)

    def _sched_switch_per_tid(self, next_tid, next_comm, prev_tid):
        # Instantiate processes if new
        self._create_proc(prev_tid)
        self._create_proc(next_tid)

        next_proc = self._state.tids[next_tid]
        next_proc.comm = next_comm
        next_proc.prev_tid = prev_tid

    def _check_prio_changed(self, timestamp, tid, prio):
        # Ignore swapper
        if tid == 0:
            return

        proc = self._state.tids[tid]

        if proc.prio != prio:
            proc.prio = prio
            self._state.send_notification_cb(
                'prio_changed', timestamp=timestamp, tid=tid, prio=prio)

    def _process_sched_switch(self, event):
        timestamp = event.timestamp
        cpu_id = event['cpu_id']
        next_tid = event['next_tid']
        next_comm = event['next_comm']
        next_prio = event['next_prio']
        prev_tid = event['prev_tid']
        prev_prio = event['prev_prio']
        prev_comm = event['prev_comm']

        self._sched_switch_per_cpu(cpu_id, next_tid)
        self._sched_switch_per_tid(next_tid, next_comm, prev_tid)
        self._check_prio_changed(timestamp, prev_tid, prev_prio)
        self._check_prio_changed(timestamp, next_tid, next_prio)

        wakee_proc = self._state.tids[next_tid]
        waker_proc = None
        if wakee_proc.last_waker is not None:
            waker_proc = self._state.tids[wakee_proc.last_waker]

        cb_data = {
            'timestamp': timestamp,
            'cpu_id': cpu_id,
            'prev_tid': prev_tid,
            'next_tid': next_tid,
            'next_comm': next_comm,
            'wakee_proc': wakee_proc,
            'waker_proc': waker_proc,
            'prev_comm': prev_comm,
        }

        self._state.send_notification_cb('sched_switch_per_cpu', **cb_data)
        self._state.send_notification_cb('sched_switch_per_tid', **cb_data)

        wakee_proc.last_wakeup = None
        wakee_proc.last_waker = None

    def _process_sched_migrate_task(self, event):
        tid = event['tid']
        prio = event['prio']

        if tid not in self._state.tids:
            proc = sv.Process()
            proc.tid = tid
            proc.comm = event['comm']
            self._state.tids[tid] = proc
        else:
            proc = self._state.tids[tid]

        self._state.send_notification_cb(
            'sched_migrate_task', proc=proc, cpu_id=event['cpu_id'])
        self._check_prio_changed(event.timestamp, tid, prio)

    def _process_sched_wakeup(self, event):
        target_cpu = event['target_cpu']
        current_cpu = event['cpu_id']
        prio = event['prio']
        tid = event['tid']

        if self._state.tracer_version < self.PRIO_OFFSET_FIX_VERSION:
            prio -= 100

        if target_cpu not in self._state.cpus:
            self._state.cpus[target_cpu] = sv.CPU(target_cpu)

        if current_cpu not in self._state.cpus:
            self._state.cpus[current_cpu] = sv.CPU(current_cpu)

        # If the TID is already executing on a CPU, ignore this wakeup
        for cpu_id in self._state.cpus:
            cpu = self._state.cpus[cpu_id]
            if cpu.current_tid == tid:
                return

        if tid not in self._state.tids:
            proc = sv.Process()
            proc.tid = tid
            self._state.tids[tid] = proc

        self._check_prio_changed(event.timestamp, tid, prio)

        # A process can be woken up multiple times, only record
        # the first one
        if self._state.tids[tid].last_wakeup is None:
            self._state.tids[tid].last_wakeup = event.timestamp
            if self._state.cpus[current_cpu].current_tid is not None:
                self._state.tids[tid].last_waker = \
                    self._state.cpus[current_cpu].current_tid

    def _process_sched_process_fork(self, event):
        child_tid = event['child_tid']
        child_pid = event['child_pid']
        child_comm = event['child_comm']
        parent_pid = event['parent_pid']
        parent_tid = event['parent_pid']
        parent_comm = event['parent_comm']

        if parent_tid not in self._state.tids:
            self._state.tids[parent_tid] = sv.Process(
                parent_tid, parent_pid, parent_comm)
        else:
            self._state.tids[parent_tid].pid = parent_pid
            self._state.tids[parent_tid].comm = parent_comm

        parent_proc = self._state.tids[parent_pid]
        child_proc = sv.Process(child_tid, child_pid, child_comm)

        for fd in parent_proc.fds:
            old_fd = parent_proc.fds[fd]
            child_proc.fds[fd] = sv.FD.new_from_fd(old_fd)
            # Note: the parent_proc key in the notification function
            # refers to the parent of the FD, which in this case is
            # the child_proc created by the fork
            self._state.send_notification_cb(
                'create_fd', fd=fd, parent_proc=child_proc,
                timestamp=event.timestamp, cpu_id=event['cpu_id'])

        self._state.tids[child_tid] = child_proc

    def _process_sched_process_exec(self, event):
        tid = event['tid']

        if tid not in self._state.tids:
            proc = sv.Process()
            proc.tid = tid
            self._state.tids[tid] = proc
        else:
            proc = self._state.tids[tid]

        # Use LTTng procname context if available
        if 'procname' in event:
            proc.comm = event['procname']

        toremove = []
        for fd in proc.fds:
            if proc.fds[fd].cloexec:
                toremove.append(fd)
        for fd in toremove:
            self._state.send_notification_cb(
                'close_fd', fd=fd, parent_proc=proc,
                timestamp=event.timestamp, cpu_id=event['cpu_id'])
            del proc.fds[fd]

    def _process_sched_pi_setprio(self, event):
        timestamp = event.timestamp
        newprio = event['newprio']
        tid = event['tid']

        self._check_prio_changed(timestamp, tid, newprio)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import period as core_period
import enum


class AnalysisConfig:
    def __init__(self):
        self.refresh_period = None
        self.begin_ts = None
        self.end_ts = None
        self.min_duration = None
        self.max_duration = None
        self.proc_list = None
        self.tid_list = None
        self.cpu_list = None
        self.period_def_registry = core_period.PeriodDefinitionRegistry()


# base class for all specific period data classes in specific analyses
class PeriodData:
    def _set_period(self, period):
        self._period = period

    @property
    def period(self):
        return self._period


@enum.unique
class AnalysisCallbackType(enum.Enum):
    TICK_CB = 'tick'


class Analysis:
    def __init__(self, state, conf, state_cbs):
        self._state = state
        self._conf = conf
        self._state_cbs = state_cbs
        self._period_key = None
        self._first_event_ts = None
        self._last_event_ts = None
        self._notification_cli_cbs = {}
        self._cbs = {}
        period_cbs = {
            core_period.PeriodEngineCallbackType.PERIOD_BEGIN:
                self._on_period_begin,
            core_period.PeriodEngineCallbackType.PERIOD_END:
                self._on_period_end,
        }
        self._period_engine = core_period.PeriodEngine(
            self._conf.period_def_registry, period_cbs)

        # This dict maps period objects (from the period module) to
        # period data objects. Period data objects are created by a
        # specific analysis implementing _create_period_data().
        self._period_data = {}

        # Mapping between a period name and it's nesting level (0 = root).
        self._period_nesting = {}

        self.started = False
        self.ended = False

    @property
    def first_event_ts(self):
        return self._first_event_ts

    @property
    def last_event_ts(self):
        return self._last_event_ts

    def period_nesting_level(self, period_name):
        if self._conf.period_def_registry.is_empty or period_name is None:
            return 0
        return self._period_nesting[period_name]

    # Returns the period data object associated with a given period.
    def _get_period_data(self, period):
        return self._period_data.get(period)

    # Sets the period data object associated with a given period.
    def _set_period_data(self, period, data):
        self._period_data[period] = data

    # Removes the period data object associated with a given period.
    def _remove_period_data(self, period):
        del self._period_data[period]

    # Creates the unique "definition-less" period. This is used when
    # there are no user-specified periods.
    def _create_defless_period(self, evt):
        period = core_period.Period(None, None, evt, None)
        self._on_period_begin(period)

    # Returns the "definition-less" period.
    def _get_defless_period(self):
        if len(self._period_data) == 0:
            return

        return next(iter(self._period_data.keys()))

    # Removes the "definition-less" period.
    def _remove_defless_period(self, completed, evt):
        period = self._get_defless_period()

        if period is None:
            return

        period.end_evt = evt
        period.completed = completed
        self._on_period_end(period)
        assert(len(self._period_data) == 0)

    # Creates a fresh specific period data object. This must be
    # implemented by a specific analysis.
    def _create_period_data(self):
        raise NotImplementedError()

    def _begin_period_cb(self, period_data):
        pass

    def _end_period_cb(self, period_data, completed,
                       begin_captures, end_captures):
        pass

    # This is called back by the period engine when a new period is
    # created. `period` is the created period, and `evt` is the event
    # that triggered the beginning of this period (the original event,
    # while `period.begin_evt` is a copy of this event).
    def _on_period_begin(self, period):
        # create the specific analysis's period data object
        period_data = self._create_period_data()

        # associate the period data object to this period object
        period_data._set_period(period)
        self._set_period_data(period, period_data)

        # register state notification callbacks with this period data object
        self._state.register_notification_cbs(period_data, self._state_cbs)

        # call specific analysis's beginning of period callback
        self._begin_period_cb(period_data)

    # This is called back by the period engine when a period is finished,
    # or closed.
    #
    # If `period.completed` is True, then the period finishes because
    # its ending expression was satisfied by an event (`period.end_evt`).
    # Otherwise, the period finishes because one of its ancestors finishes,
    # or because the period engine user asked for it.
    def _on_period_end(self, period):
        # get the period data object associated with this period object
        period_data = self._get_period_data(period)

        # call specific analysis's end of period callback
        self._end_period_cb(period_data, period.completed,
                            period.begin_captures, period.end_captures)

        # send tick notification to owner (CLI)
        self._send_notification_cb(AnalysisCallbackType.TICK_CB, period_data,
                                   end_ns=self.last_event_ts)

        # clear registered state notification callbacks associated with
        # this period
        self._state.clear_period_notification_cbs(period_data)

        # remove this period data object
        self._remove_period_data(period)

    # This is called by the owner of this analysis when an event must
    # be processed (`ev`).
    def process_event(self, ev):
        self._check_analysis_end(ev)
        if self.ended:
            return

        if self._first_event_ts is None:
            self._first_event_ts = ev.timestamp

        self._last_event_ts = ev.timestamp

        if not self.started:
            if self._conf.begin_ts:
                self._check_analysis_begin(ev)
                if not self.started:
                    return
            else:
                self.started = True

        # Run the period engine. This call has the effect of calling
        # back _on_period_begin() or _on_period_end(), zero or more
        # times, for each beginning and ending period according to the
        # registered period definitions.
        self._period_engine.process_event(ev)

        # check the refresh period conditions
        self._check_refresh(ev)

    # Create the mapping between a period name and its nesting level.
    # Recursively iterate over all children.
    def _get_period_nesting_level(self, period_def, level):
        for child in period_def.children:
            self._get_period_nesting_level(child, level + 1)
        self._period_nesting[period_def.name] = level

    # Iterate over all root period definitions to create the map between
    # a period name and it's nesting level.
    def _create_period_nesting_map(self):
        for period_def in self._conf.period_def_registry._root_period_defs:
            self._get_period_nesting_level(period_def, 0)

    # Called by the owner of this analysis to indicate that this
    # analysis is starting.
    def begin_analysis(self, evt):
        # If we do not have any period defined, create the
        # "definition-less" period starting at the first event.
        if (self._conf.period_def_registry.is_empty and
                self._conf.begin_ts is None):
            self._create_defless_period(evt)
        self._create_period_nesting_map()

    def end_analysis(self):
        # let the periods know that it is the last one
        self.ended = True

        # This is the end of the analysis, so we need to remove all
        # the existing periods. This means either remove all the existing
        # periods in the period engine, or remove the unique,
        # "definition-less" period created here.
        if self._conf.period_def_registry.is_empty:
            self._remove_defless_period(False, None)
        else:
            self._period_engine.remove_all_periods()
            self._period_data.clear()

        # Send an empty TICK notification if the CLI needs to
        # do something at the end even if there are no existing
        # periods.
        self._send_notification_cb(AnalysisCallbackType.TICK_CB, None,
                                   end_ns=self._last_event_ts)

    def register_notification_cbs(self, cbs):
        for name in cbs:
            if name not in self._notification_cli_cbs:
                self._notification_cli_cbs[name] = []

            self._notification_cli_cbs[name].append(cbs[name])

    def _send_notification_cb(self, name, period, **kwargs):
        if name in self._notification_cli_cbs:
            for cb in self._notification_cli_cbs[name]:
                cb(period, **kwargs)

    def _register_cbs(self, cbs):
        self._cbs = cbs

    def _process_event_cb(self, ev):
        name = ev.name

        if name in self._cbs:
            self._cbs[name](ev)
        elif 'syscall_entry' in self._cbs and \
             (name.startswith('sys_') or name.startswith('syscall_entry_')):
            self._cbs['syscall_entry'](ev)
        elif 'syscall_exit' in self._cbs and \
                (name.startswith('exit_syscall') or
                 name.startswith('syscall_exit_')):
            self._cbs['syscall_exit'](ev)

    def _check_analysis_begin(self, ev):
        if self._conf.begin_ts and ev.timestamp >= self._conf.begin_ts:
            self._create_defless_period(ev)
            self.started = True

    def _check_analysis_end(self, ev):
        if self._conf.end_ts and ev.timestamp > self._conf.end_ts:
            self.ended = True

    def _check_refresh(self, evt):
        if self._conf.refresh_period is None:
            return

        period = self._get_defless_period()

        if evt.timestamp >= (period.begin_evt.timestamp +
                             self._conf.refresh_period):
            # remove the current period and create a new one
            self._remove_defless_period(True, evt)
            self._create_defless_period(evt)

    def _filter_process(self, proc):
        if not proc:
            return True
        if self._conf.proc_list and proc.comm not in self._conf.proc_list:
            return False
        if self._conf.tid_list and proc.tid not in self._conf.tid_list:
            return False
        return True

    def _filter_cpu(self, cpu):
        return not (self._conf.cpu_list and cpu not in self._conf.cpu_list)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#                      Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest
from lttnganalyses.common import trace_utils
from .analysis_test import AnalysisTest


class IntersectTest(AnalysisTest):
    def write_trace(self):
        # Write these events in the default stream.
        self.trace_writer.write_softirq_raise(1005, 3, 1)
        self.trace_writer.write_softirq_entry(1006, 3, 1)
        self.trace_writer.write_softirq_exit(1009, 3, 1)

        # Override the default stream, so all new events are written
        # in a different stream, no overlapping timestamps between streams.
        self.trace_writer.create_stream()
        self.trace_writer.write_softirq_exit(1010, 2, 7)
        self.trace_writer.flush()

    @unittest.skipIf(trace_utils.read_babeltrace_version() <
                     trace_utils.BT_INTERSECT_VERSION,
                     "not supported by Babeltrace < %s" %
                     trace_utils.BT_INTERSECT_VERSION,)
    def test_no_intersection(self):
        test_name = 'no_intersection'
        expected = self.get_expected_output(test_name)
        result = self.get_cmd_output('lttng-irqstats')

        self._assertMultiLineEqual(result, expected, test_name)

    def test_disable_intersect(self):
        test_name = 'disable_intersect'
        expected = self.get_expected_output(test_name)
        result = self.get_cmd_output('lttng-irqstats',
                                     options='--no-intersection')

        self._assertMultiLineEqual(result, expected, test_name)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import sp


class MemStateProvider(sp.StateProvider):
    def __init__(self, state):
        cbs = {
            'mm_page_alloc': self._process_mm_page_alloc,
            'kmem_mm_page_alloc': self._process_mm_page_alloc,
            'mm_page_free': self._process_mm_page_free,
            'kmem_mm_page_free': self._process_mm_page_free,
        }

        super().__init__(state, cbs)

    def _get_current_proc(self, event):
        cpu_id = event['cpu_id']
        if cpu_id not in self._state.cpus:
            return None

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return None

        return self._state.tids[cpu.current_tid]

    def _process_mm_page_alloc(self, event):
        self._state.mm.page_count += 1

        # Increment the number of pages allocated during the execution
        # of all currently syscall io requests
        for process in self._state.tids.values():
            if process.current_syscall is None:
                continue

            if process.current_syscall.io_rq:
                process.current_syscall.io_rq.pages_allocated += 1

        current_process = self._get_current_proc(event)
        if current_process is None:
            return

        self._state.send_notification_cb('tid_page_alloc',
                                         proc=current_process,
                                         cpu_id=event['cpu_id'])

    def _process_mm_page_free(self, event):
        if self._state.mm.page_count == 0:
            return

        self._state.mm.page_count -= 1

        current_process = self._get_current_proc(event)
        if current_process is None:
            return

        self._state.send_notification_cb('tid_page_free',
                                         proc=current_process,
                                         cpu_id=event['cpu_id'])



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import socket
from babeltrace import CTFScope
from . import sp, sv
from ..common import format_utils, trace_utils


class IoStateProvider(sp.StateProvider):
    def __init__(self, state):
        cbs = {
            'syscall_entry': self._process_syscall_entry,
            'syscall_exit': self._process_syscall_exit,
            'syscall_entry_connect': self._process_connect,
            'writeback_pages_written': self._process_writeback_pages_written,
            'mm_vmscan_wakeup_kswapd': self._process_mm_vmscan_wakeup_kswapd,
            'mm_page_free': self._process_mm_page_free
        }

        super().__init__(state, cbs)

    def _process_syscall_entry(self, event):
        # Only handle IO Syscalls
        name = trace_utils.get_syscall_name(event)
        if name not in sv.SyscallConsts.IO_SYSCALLS:
            return

        cpu_id = event['cpu_id']
        if cpu_id not in self._state.cpus:
            return

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return

        proc = self._state.tids[cpu.current_tid]

        # check if we can fix the pid from a context
        self._fix_context_pid(event, proc)

        if name in sv.SyscallConsts.OPEN_SYSCALLS:
            self._track_open(event, name, proc)
        elif name in sv.SyscallConsts.CLOSE_SYSCALLS:
            self._track_close(event, name, proc)
        elif name in sv.SyscallConsts.READ_SYSCALLS or \
                name in sv.SyscallConsts.WRITE_SYSCALLS:
            self._track_read_write(event, name, proc)
        elif name in sv.SyscallConsts.SYNC_SYSCALLS:
            self._track_sync(event, name, proc)

    def _process_syscall_exit(self, event):
        cpu_id = event['cpu_id']
        if cpu_id not in self._state.cpus:
            return

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return

        proc = self._state.tids[cpu.current_tid]
        current_syscall = proc.current_syscall
        if current_syscall is None:
            return

        name = current_syscall.name
        if name not in sv.SyscallConsts.IO_SYSCALLS:
            return

        self._track_io_rq_exit(event, proc)

        proc.current_syscall = None

    def _process_connect(self, event):
        cpu_id = event['cpu_id']
        if cpu_id not in self._state.cpus:
            return

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return

        proc = self._state.tids[cpu.current_tid]
        parent_proc = self._get_parent_proc(proc)

        # FIXME: handle on syscall_exit_connect only when succesful
        if 'family' in event and event['family'] == socket.AF_INET:
            fd = event['fd']
            if fd in parent_proc.fds:
                parent_proc.fds[fd].filename = format_utils.format_ipv4(
                    event['v4addr'], event['dport']
                )
            self._state.send_notification_cb('update_fd',
                                             fd=fd,
                                             parent_proc=proc,
                                             timestamp=event.timestamp,
                                             cpu_id=event['cpu_id'])

    def _process_writeback_pages_written(self, event):
        for cpu in self._state.cpus.values():
            if cpu.current_tid is None:
                continue

            current_syscall = self._state.tids[cpu.current_tid].current_syscall
            if current_syscall is None:
                continue

            if current_syscall.io_rq:
                current_syscall.io_rq.pages_written += event['pages']

    def _process_mm_vmscan_wakeup_kswapd(self, event):
        cpu_id = event['cpu_id']
        if cpu_id not in self._state.cpus:
            return

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return

        current_syscall = self._state.tids[cpu.current_tid].current_syscall
        if current_syscall is None:
            return

        if current_syscall.io_rq:
            current_syscall.io_rq.woke_kswapd = True

    def _process_mm_page_free(self, event):
        for cpu in self._state.cpus.values():
            if cpu.current_tid is None:
                continue

            proc = self._state.tids[cpu.current_tid]

            # if the current process is kswapd0, we need to
            # attribute the page freed to the process that
            # woke it up.
            if proc.comm == 'kswapd0' and proc.prev_tid > 0:
                proc = self._state.tids[proc.prev_tid]

            current_syscall = proc.current_syscall
            if current_syscall is None:
                continue

            if current_syscall.io_rq and current_syscall.io_rq.woke_kswapd:
                current_syscall.io_rq.pages_freed += 1

    def _track_open(self, event, name, proc):
        current_syscall = proc.current_syscall
        if name in sv.SyscallConsts.DISK_OPEN_SYSCALLS:
            current_syscall.io_rq = sv.OpenIORequest.new_from_disk_open(
                event, proc.tid)
        elif name in ['accept', 'accept4']:
            current_syscall.io_rq = sv.OpenIORequest.new_from_accept(
                event, proc.tid)
        elif name == 'socket':
            current_syscall.io_rq = sv.OpenIORequest.new_from_socket(
                event, proc.tid)
        elif name in sv.SyscallConsts.DUP_OPEN_SYSCALLS:
            self._track_dup(event, name, proc)

    def _track_dup(self, event, name, proc):
        current_syscall = proc.current_syscall

        # If the process that triggered the io_rq is a thread,
        # its FDs are that of the parent process
        parent_proc = self._get_parent_proc(proc)
        fds = parent_proc.fds

        if name == 'dup':
            oldfd = event['fildes']
        elif name in ['dup2', 'dup3']:
            oldfd = event['oldfd']
            newfd = event['newfd']
            if newfd in fds:
                self._close_fd(parent_proc, newfd, event.timestamp,
                               event['cpu_id'])
        elif name == 'fcntl':
            # Only handle if cmd == F_DUPFD (0)
            if event['cmd'] != 0:
                return

            oldfd = event['fd']

        old_file = None
        if oldfd in fds:
            old_file = fds[oldfd]

        current_syscall.io_rq = sv.OpenIORequest.new_from_old_fd(
            event, proc.tid, old_file)

        if name == 'dup3':
            cloexec = event['flags'] & os.O_CLOEXEC == os.O_CLOEXEC
            current_syscall.io_rq.cloexec = cloexec

    def _track_close(self, event, name, proc):
        proc.current_syscall.io_rq = sv.CloseIORequest(
            event.timestamp, proc.tid, event['fd'])

    def _track_read_write(self, event, name, proc):
        current_syscall = proc.current_syscall

        if name == 'splice':
            current_syscall.io_rq = sv.ReadWriteIORequest.new_from_splice(
                event, proc.tid)
            return
        elif name == 'sendfile64':
            current_syscall.io_rq = sv.ReadWriteIORequest.new_from_sendfile64(
                event, proc.tid)
            return

        if name in ['writev', 'pwritev', 'readv', 'preadv']:
            size_key = 'vlen'
        elif name == 'recvfrom':
            size_key = 'size'
        elif name == 'sendto':
            size_key = 'len'
        elif name in ['recvmsg', 'sendmsg']:
            size_key = None
        else:
            size_key = 'count'

        current_syscall.io_rq = sv.ReadWriteIORequest.new_from_fd_event(
            event, proc.tid, size_key)

    def _track_sync(self, event, name, proc):
        current_syscall = proc.current_syscall

        if name == 'sync':
            current_syscall.io_rq = sv.SyncIORequest.new_from_sync(
                event, proc.tid)
        elif name in ['fsync', 'fdatasync']:
            current_syscall.io_rq = sv.SyncIORequest.new_from_fsync(
                event, proc.tid)
        elif name == 'sync_file_range':
            current_syscall.io_rq = sv.SyncIORequest.new_from_sync_file_range(
                event, proc.tid)

    def _track_io_rq_exit(self, event, proc):
        ret = event['ret']
        cpu_id = event['cpu_id']
        io_rq = proc.current_syscall.io_rq
        # io_rq can be None in the case of fcntl when cmd is not
        # F_DUPFD, in which case we disregard the syscall as it did
        # not open any FD
        if io_rq is None:
            return

        io_rq.update_from_exit(event)

        if ret >= 0:
            self._create_fd(proc, io_rq, cpu_id)

        parent_proc = self._get_parent_proc(proc)
        self._state.send_notification_cb('io_rq_exit',
                                         io_rq=io_rq,
                                         proc=proc,
                                         parent_proc=parent_proc,
                                         cpu_id=cpu_id)

        if isinstance(io_rq, sv.CloseIORequest) and ret == 0:
            self._close_fd(proc, io_rq.fd, io_rq.end_ts, cpu_id)

    def _create_fd(self, proc, io_rq, cpu_id):
        parent_proc = self._get_parent_proc(proc)

        if io_rq.fd is not None and io_rq.fd not in parent_proc.fds:
            if isinstance(io_rq, sv.OpenIORequest):
                parent_proc.fds[io_rq.fd] = sv.FD.new_from_open_rq(io_rq)
            else:
                parent_proc.fds[io_rq.fd] = sv.FD(io_rq.fd)

            self._state.send_notification_cb('create_fd',
                                             fd=io_rq.fd,
                                             parent_proc=parent_proc,
                                             timestamp=io_rq.end_ts,
                                             cpu_id=cpu_id)
        elif isinstance(io_rq, sv.ReadWriteIORequest):
            if io_rq.fd_in is not None and io_rq.fd_in not in parent_proc.fds:
                parent_proc.fds[io_rq.fd_in] = sv.FD(io_rq.fd_in)
                self._state.send_notification_cb('create_fd',
                                                 fd=io_rq.fd_in,
                                                 parent_proc=parent_proc,
                                                 timestamp=io_rq.end_ts,
                                                 cpu_id=cpu_id)

            if io_rq.fd_out is not None and \
               io_rq.fd_out not in parent_proc.fds:
                parent_proc.fds[io_rq.fd_out] = sv.FD(io_rq.fd_out)
                self._state.send_notification_cb('create_fd',
                                                 fd=io_rq.fd_out,
                                                 parent_proc=parent_proc,
                                                 timestamp=io_rq.end_ts,
                                                 cpu_id=cpu_id)

    def _close_fd(self, proc, fd, timestamp, cpu_id):
        parent_proc = self._get_parent_proc(proc)
        self._state.send_notification_cb('close_fd',
                                         fd=fd,
                                         parent_proc=parent_proc,
                                         timestamp=timestamp,
                                         cpu_id=cpu_id)
        del parent_proc.fds[fd]

    def _get_parent_proc(self, proc):
        if proc.pid is not None and proc.tid != proc.pid:
            parent_proc = self._state.tids[proc.pid]
        else:
            parent_proc = proc

        return parent_proc

    def _fix_context_pid(self, event, proc):
        for context in event.field_list_with_scope(
                CTFScope.STREAM_EVENT_CONTEXT):
            if context != 'pid':
                continue
            # make sure the 'pid' field is not also in the event
            # payload, otherwise we might clash
            for context in event.field_list_with_scope(
                    CTFScope.EVENT_FIELDS):
                if context == 'pid':
                    return

            if proc.pid is None:
                proc.pid = event['pid']
                if event['pid'] != proc.tid:
                    proc.pid = event['pid']
                    parent_proc = sv.Process(proc.pid, proc.pid, proc.comm,
                                             proc.prio)
                    self._state.tids[parent_proc.pid] = parent_proc



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import stats
from .analysis import Analysis, PeriodData


class _PeriodData(PeriodData):
    def __init__(self):
        # Log of individual wake scheduling events
        self.sched_list = []
        self.min_latency = None
        self.max_latency = None
        self.total_latency = 0
        self.tids = {}


class SchedAnalysis(Analysis):
    def __init__(self, state, conf):
        notification_cbs = {
            'sched_switch_per_tid': self._process_sched_switch,
            'prio_changed': self._process_prio_changed,
        }
        super().__init__(state, conf, notification_cbs)

    def count(self, period_data):
        return len(period_data.sched_list)

    def _create_period_data(self):
        return _PeriodData()

    def _process_sched_switch(self, period_data, **kwargs):
        cpu_id = kwargs['cpu_id']
        switch_ts = kwargs['timestamp']
        wakee_proc = kwargs['wakee_proc']
        waker_proc = kwargs['waker_proc']
        next_tid = kwargs['next_tid']
        wakeup_ts = wakee_proc.last_wakeup
#        print(period_data)

        if not self._filter_process(wakee_proc):
            return
        if not self._filter_cpu(cpu_id):
            return

        if wakeup_ts is None:
            return

        latency = switch_ts - wakeup_ts
        if self._conf.min_duration is not None and \
           latency < self._conf.min_duration:
            return
        if self._conf.max_duration is not None and \
           latency > self._conf.max_duration:
            return

        if waker_proc is not None and waker_proc.tid not in period_data.tids:
            period_data.tids[waker_proc.tid] = \
                ProcessSchedStats.new_from_process(waker_proc)
            period_data.tids[waker_proc.tid].update_prio(switch_ts,
                                                         waker_proc.prio)

        if next_tid not in period_data.tids:
            period_data.tids[next_tid] = \
                ProcessSchedStats.new_from_process(wakee_proc)
            period_data.tids[next_tid].update_prio(switch_ts, wakee_proc.prio)

        sched_event = SchedEvent(
            wakeup_ts, switch_ts, wakee_proc, waker_proc, cpu_id)
        period_data.tids[next_tid].update_stats(sched_event)
        self._update_stats(period_data, sched_event)

    def _process_prio_changed(self, period_data, **kwargs):
        timestamp = kwargs['timestamp']
        prio = kwargs['prio']
        tid = kwargs['tid']

        if tid not in period_data.tids:
            return

        period_data.tids[tid].update_prio(timestamp, prio)

    def _update_stats(self, period_data, sched_event):
        if period_data.min_latency is None or \
                sched_event.latency < period_data.min_latency:
            period_data.min_latency = sched_event.latency

        if period_data.max_latency is None or \
                sched_event.latency > period_data.max_latency:
            period_data.max_latency = sched_event.latency

        period_data.total_latency += sched_event.latency
        period_data.sched_list.append(sched_event)


class ProcessSchedStats(stats.Process):
    def __init__(self, pid, tid, comm):
        super().__init__(pid, tid, comm)

        self.min_latency = None
        self.max_latency = None
        self.total_latency = 0
        self.sched_list = []

    @property
    def count(self):
        return len(self.sched_list)

    def update_stats(self, sched_event):
        if self.min_latency is None or sched_event.latency < self.min_latency:
            self.min_latency = sched_event.latency

        if self.max_latency is None or sched_event.latency > self.max_latency:
            self.max_latency = sched_event.latency

        self.total_latency += sched_event.latency
        self.sched_list.append(sched_event)

    def reset(self):
        super().reset()
        self.min_latency = None
        self.max_latency = None
        self.total_latency = 0
        self.sched_list = []


class SchedEvent():
    def __init__(self, wakeup_ts, switch_ts, wakee_proc, waker_proc,
                 target_cpu):
        self.wakeup_ts = wakeup_ts
        self.switch_ts = switch_ts
        self.wakee_proc = wakee_proc
        self.waker_proc = waker_proc
        self.prio = wakee_proc.prio
        self.target_cpu = target_cpu
        self.latency = switch_ts - wakeup_ts



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import stats
from .analysis import Analysis, PeriodData


class _PeriodData(PeriodData):
    def __init__(self):
        self.period_begin_ts = None
        self.cpus = {}
        self.tids = {}


class Cputop(Analysis):
    def __init__(self, state, conf):
        notification_cbs = {
            'sched_migrate_task': self._process_sched_migrate_task,
            'sched_switch_per_cpu': self._process_sched_switch_per_cpu,
            'sched_switch_per_tid': self._process_sched_switch_per_tid,
            'prio_changed': self._process_prio_changed,
        }

        super().__init__(state, conf, notification_cbs)

    def _create_period_data(self):
        return _PeriodData()

    def _begin_period_cb(self, period_data):
        period = period_data.period
        period_data.period_begin_ts = period.begin_evt.timestamp

    def _end_period_cb(self, period_data, completed, begin_captures,
                       end_captures):
        self._compute_stats(period_data)

    def _compute_stats(self, period_data):
        """Compute usage stats relative to a certain time range

        For each CPU and process tracked by the analysis, we set its
        usage_percent attribute, which represents the percentage of
        usage time for the given CPU or process relative to the full
        duration of the time range. Do note that we need to know the
        timestamps and not just the duration, because if a CPU or a
        process is currently busy, we use the end timestamp to add
        the partial results of the currently running task to the usage
        stats.
        """
        duration = self.last_event_ts - period_data.period.begin_evt.timestamp

        for cpu_id in period_data.cpus:
            cpu = period_data.cpus[cpu_id]
            if cpu.current_task_start_ts is not None:
                cpu.total_usage_time += self.last_event_ts - \
                    cpu.current_task_start_ts

            cpu.compute_stats(duration)

        for tid in period_data.tids:
            proc = period_data.tids[tid]
            if proc.last_sched_ts is not None:
                proc.total_cpu_time += self.last_event_ts - \
                    proc.last_sched_ts

            proc.compute_stats(duration)

    def _process_sched_switch_per_cpu(self, period_data, **kwargs):
        timestamp = kwargs['timestamp']
        cpu_id = kwargs['cpu_id']
        wakee_proc = kwargs['wakee_proc']

        if not self._filter_cpu(cpu_id):
            return

        if cpu_id not in period_data.cpus:
            period_data.cpus[cpu_id] = CpuUsageStats(cpu_id)
            period_data.cpus[cpu_id].current_task_start_ts = \
                period_data.period_begin_ts

        cpu = period_data.cpus[cpu_id]
        if cpu.current_task_start_ts is not None:
            cpu.total_usage_time += timestamp - cpu.current_task_start_ts

        if not self._filter_process(wakee_proc):
            cpu.current_task_start_ts = None
        else:
            cpu.current_task_start_ts = timestamp

    def _process_sched_switch_per_tid(self, period_data, **kwargs):
        cpu_id = kwargs['cpu_id']
        wakee_proc = kwargs['wakee_proc']
        timestamp = kwargs['timestamp']
        prev_tid = kwargs['prev_tid']
        next_tid = kwargs['next_tid']
        next_comm = kwargs['next_comm']
        prev_comm = kwargs['prev_comm']

        if not self._filter_cpu(cpu_id):
            return

        if prev_tid not in period_data.tids:
            period_data.tids[prev_tid] = ProcessCpuStats(
                None, prev_tid, prev_comm)
            prev_proc = period_data.tids[prev_tid]
            # Set the last_sched_ts to the beginning of the period
            # since we missed the entry event.
            prev_proc.last_sched_ts = period_data.period_begin_ts

        prev_proc = period_data.tids[prev_tid]
        if prev_proc.last_sched_ts is not None:
            prev_proc.total_cpu_time += timestamp - prev_proc.last_sched_ts
            prev_proc.last_sched_ts = None

        # Only filter on wakee_proc after finalizing the prev_proc
        # accounting
        if not self._filter_process(wakee_proc):
            return

        if next_tid not in period_data.tids:
            period_data.tids[next_tid] = ProcessCpuStats(None,
                                                         next_tid, next_comm)
            period_data.tids[next_tid].update_prio(timestamp, wakee_proc.prio)

        next_proc = period_data.tids[next_tid]
        next_proc.last_sched_ts = timestamp

    def _process_sched_migrate_task(self, period_data, **kwargs):
        cpu_id = kwargs['cpu_id']
        proc = kwargs['proc']
        tid = proc.tid

        if not self._filter_process(proc):
            return
        if not self._filter_cpu(cpu_id):
            return

        if tid not in period_data.tids:
            period_data.tids[tid] = ProcessCpuStats.new_from_process(proc)

        period_data.tids[tid].migrate_count += 1

    def _process_prio_changed(self, period_data, **kwargs):
        timestamp = kwargs['timestamp']
        prio = kwargs['prio']
        tid = kwargs['tid']

        if tid not in period_data.tids:
            return

        period_data.tids[tid].update_prio(timestamp, prio)

    def _filter_process(self, proc):
        # Exclude swapper
        if proc.tid == 0:
            return False

        return super()._filter_process(proc)


class CpuUsageStats():
    def __init__(self, cpu_id):
        self.cpu_id = cpu_id
        # Usage time and start timestamp are in nanoseconds (ns)
        self.total_usage_time = 0
        self.current_task_start_ts = None
        self.usage_percent = None

    def compute_stats(self, duration):
        if duration != 0:
            self.usage_percent = self.total_usage_time * 100 / duration
        else:
            self.usage_percent = 0

    def reset(self):
        self.total_usage_time = 0
        self.usage_percent = None


class ProcessCpuStats(stats.Process):
    def __init__(self, pid, tid, comm):
        super().__init__(pid, tid, comm)

        # CPU Time and timestamp in nanoseconds (ns)
        self.total_cpu_time = 0
        self.last_sched_ts = None
        self.migrate_count = 0
        self.usage_percent = None

    def compute_stats(self, duration):
        if duration != 0:
            self.usage_percent = self.total_cpu_time * 100 / duration
        else:
            self.usage_percent = 0

    def reset(self):
        super().reset()
        self.total_cpu_time = 0
        self.migrate_count = 0
        self.usage_percent = None



#!/usr/bin/env python3
#
# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Helper tool to generate CTFWriter code from the metadata of an existing
# trace.
# It used to add code in TraceTest.py.
# Only the basic types are supported, a warning is generated if a field cannot
# be generated so it is easy to look manually at the metadata and fix it.

import sys
import argparse

from babeltrace import TraceCollection, CTFScope, CTFTypeId


def sanitize(s):
    """Replace special characters in s by underscores.

    This makes s suitable to use in code as a function or variable name.
    """
    s = s.replace(':', '_')

    return s


def get_definition_type(field, event):
    event_name = sanitize(event.name)

    if field.type == CTFTypeId.INTEGER:
        signed = ''
        if field.signedness == 0:
            signed = 'u'
        length = field.length
        print('        self.%s.add_field(self.%sint%s_type, "_%s")' %
              (event_name, signed, length, field.name))
    elif field.type == CTFTypeId.ARRAY:
        print('        self.%s.add_field(self.array%s_type, "_%s")' %
              (event_name, field.length, field.name))
    elif field.type == CTFTypeId.STRING:
        print('        self.%s.add_field(self.string_type, "_%s")' %
              (event_name, field.name))
    else:
        print('        # FIXME %s.%s: Unhandled type %d' % (event.name,
                                                            field.name,
                                                            field.type))


def gen_define(event):
        fields = []
        event_name = sanitize(event.name)
        print('    def define_%s(self):' % (event_name))
        print('        self.%s = CTFWriter.EventClass("%s")' %
              (event_name, event.name))
        for field in event.fields:
            if field.scope == CTFScope.EVENT_FIELDS:
                fname = field.name
                fields.append(fname)
                get_definition_type(field, event)
        print('        self.add_event(self.%s)' % event_name)
        print('')
        return fields


def gen_write(event, fields):
        f_list = ''
        for f in fields:
            f_list += ', {}'.format(f)

        event_name = sanitize(event.name)
        print('    def write_%s(self, time_ms, cpu_id%s):' % (event_name,
                                                              f_list))
        print('        event = CTFWriter.Event(self.%s)' % (event_name))
        print('        self.clock.time = time_ms * 1000000')
        print('        self.set_int(event.payload("_cpu_id"), cpu_id)')
        for field in event.fields:
            if field.scope == CTFScope.EVENT_FIELDS:
                fname = field.name
                if field.type == CTFTypeId.INTEGER:
                    print('        self.set_int(event.payload("_%s"), %s)' %
                          (fname, fname))
                elif field.type == CTFTypeId.ARRAY:
                    print('        self.set_char_array(event.payload("_%s"), '
                          '%s)' % (fname, fname))
                elif field.type == CTFTypeId.STRING:
                    print('        self.set_string(event.payload("_%s"), %s)' %
                          (fname, fname))
                else:
                    print('        # FIXME %s.%s: Unhandled type %d' %
                          (event.name, field.name, field.type))
        print('        self.stream.append_event(event)')
        print('        self.stream.flush()')
        print('')


def gen_parser(handle, args):
    for h in handle.values():
        for event in h.events:
            fields = gen_define(event)
            gen_write(event, fields)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='CTFWriter code generator')
    parser.add_argument('path', metavar="<path/to/trace>", help='Trace path')
    args = parser.parse_args()

    traces = TraceCollection()
    handle = traces.add_traces_recursive(args.path, "ctf")
    if handle is None:
        sys.exit(1)

    gen_parser(handle, args)

    for h in handle.values():
        traces.remove_trace(h)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from . import sp, sv


class StatedumpStateProvider(sp.StateProvider):
    def __init__(self, state):
        cbs = {
            'lttng_statedump_process_state':
            self._process_lttng_statedump_process_state,
            'lttng_statedump_file_descriptor':
            self._process_lttng_statedump_file_descriptor,
            'lttng_statedump_block_device':
            self._process_lttng_statedump_block_device
        }

        super().__init__(state, cbs)

    def _process_lttng_statedump_block_device(self, event):
        dev = event['dev']
        diskname = event['diskname']

        if dev not in self._state.disks:
            self._state.disks[dev] = sv.Disk(dev, diskname=diskname)
        elif self._state.disks[dev].diskname is None:
            self._state.disks[dev].diskname = diskname
        self._state.send_notification_cb('lttng_statedump_block_device',
                                         dev=dev, diskname=diskname)

    def _process_lttng_statedump_process_state(self, event):
        tid = event['tid']
        pid = event['pid']
        name = event['name']
        # prio is not in the payload for LTTng-modules < 2.8. Using
        # get() will set it to None if the key is not found
        prio = event.get('prio')

        if tid not in self._state.tids:
            self._state.tids[tid] = sv.Process(tid=tid)

        proc = self._state.tids[tid]
        # Even if the process got created earlier, some info might be
        # missing, add it now.
        proc.pid = pid
        proc.comm = name
        # However don't override the prio value if we already got the
        # information from sched_* events.
        if proc.prio is None:
            proc.prio = prio

        if pid != tid:
            # create the parent
            if pid not in self._state.tids:
                # FIXME: why is the parent's name set to that of the
                # child? does that make sense?

                # tid == pid for the parent process
                self._state.tids[pid] = sv.Process(tid=pid, pid=pid, comm=name)

            parent = self._state.tids[pid]
            # If the thread had opened FDs, they need to be assigned
            # to the parent.
            StatedumpStateProvider._assign_fds_to_parent(proc, parent)
            self._state.send_notification_cb('create_parent_proc',
                                             proc=proc,
                                             parent_proc=parent)

    def _process_lttng_statedump_file_descriptor(self, event):
        pid = event['pid']
        fd = event['fd']
        filename = event['filename']
        cloexec = event['flags'] & os.O_CLOEXEC == os.O_CLOEXEC

        if pid not in self._state.tids:
            self._state.tids[pid] = sv.Process(tid=pid, pid=pid)

        proc = self._state.tids[pid]

        if fd not in proc.fds:
            proc.fds[fd] = sv.FD(fd, filename, sv.FDType.unknown, cloexec)
            self._state.send_notification_cb('create_fd',
                                             fd=fd,
                                             parent_proc=proc,
                                             timestamp=event.timestamp,
                                             cpu_id=event['cpu_id'])
        else:
            # just fix the filename
            proc.fds[fd].filename = filename
            self._state.send_notification_cb('update_fd',
                                             fd=fd,
                                             parent_proc=proc,
                                             timestamp=event.timestamp,
                                             cpu_id=event['cpu_id'])

    @staticmethod
    def _assign_fds_to_parent(proc, parent):
        if proc.fds:
            toremove = []
            for fd in proc.fds:
                if fd not in parent.fds:
                    parent.fds[fd] = proc.fds[fd]
                else:
                    # best effort to fix the filename
                    if not parent.fds[fd].filename:
                        parent.fds[fd].filename = proc.fds[fd].filename
                toremove.append(fd)
            for fd in toremove:
                del proc.fds[fd]



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import datetime
import re
from time import timezone
from . import trace_utils
from .time_utils import NSEC_PER_SEC


def _split_value_units(raw_str):
    """Take a string with a numerical value and units, and separate the
    two.

    Args:
        raw_str (str): the string to parse, with numerical value and
        (optionally) units.

    Returns:
        A tuple (value, units), where value is a string and units is
        either a string or `None` if no units were found.
    """
    try:
        units_index = next(i for i, c in enumerate(raw_str) if c.isalpha())
    except StopIteration:
        # no units found
        return (raw_str, None)

    return (raw_str[:units_index], raw_str[units_index:])


def parse_size(size_str):
    """Convert a human-readable size string to an integral number of
    bytes.

    Args:
        size_str (str): the formatted string comprised of the size and
        units.

    Returns:
        A number of bytes.

    Raises:
        ValueError: if units are unrecognised or the size is not a
        real number.
    """
    binary_units = ['B', 'KiB', 'MiB', 'GiB', 'TiB',
                    'PiB', 'EiB', 'ZiB', 'YiB']
    # units as printed by GNU coreutils (e.g. ls or du), using base
    # 1024 as well
    coreutils_units = ['B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
    si_units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']

    size, units = _split_value_units(size_str)

    try:
        size = float(size)
    except ValueError:
        raise ValueError('invalid size: {}'.format(size))

    # If no units have been found, assume bytes
    if units is not None:
        if units in binary_units:
            base = 1024
            exponent = binary_units.index(units)
        elif units in coreutils_units:
            base = 1024
            exponent = coreutils_units.index(units)
        elif units in si_units:
            base = 1000
            exponent = si_units.index(units)
        else:
            raise ValueError('unrecognised units: {}'.format(units))

        size *= base ** exponent

    return int(size)


def parse_duration(duration_str):
    """Convert a human-readable duration string to an integral number of
    nanoseconds.

    Args:
        duration_str (str): the formatted string comprised of the
        duration and units.

    Returns:
        A number of nanoseconds.

    Raises:
        ValueError: if units are unrecognised or the size is not a
        real number.
    """
    base = 1000
    duration, units = _split_value_units(duration_str)

    try:
        duration = float(duration)
    except ValueError:
        raise ValueError('invalid duration: {}'.format(duration))

    if units is not None:
        if units == 's':
            exponent = 3
        elif units == 'ms':
            exponent = 2
        elif units in ['us', 'µs']:
            exponent = 1
        elif units == 'ns':
            exponent = 0
        else:
            raise ValueError('unrecognised units: {}'.format(units))
    else:
        # no units defaults to seconds
        exponent = 3

    duration *= base ** exponent

    return int(duration)


def _parse_date_full_with_nsec(date):
    """Parse full date string with nanosecond resolution.

    This matches either 2014-12-12 17:29:43.802588035 or
    2014-12-12T17:29:43.802588035.

    Args:
        date (str): the date string to be parsed.

    Returns:
        A tuple of the format (date_time, nsec), where date_time is a
        datetime.datetime object and nsec is an int of the remaining
        nanoseconds.

    Raises:
        ValueError: if the date format does not match.
    """
    pattern = re.compile(
        r'^(?P<year>\d{4})-(?P<mon>[01]\d)-(?P<day>[0-3]\d)[\sTt]'
        r'(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<nsec>\d{9})$'
    )

    if not pattern.match(date):
        raise ValueError('Wrong date format: {}'.format(date))

    year = pattern.search(date).group('year')
    month = pattern.search(date).group('mon')
    day = pattern.search(date).group('day')
    hour = pattern.search(date).group('hour')
    minute = pattern.search(date).group('min')
    sec = pattern.search(date).group('sec')
    nsec = pattern.search(date).group('nsec')

    date_time = datetime.datetime(
        int(year), int(month), int(day),
        int(hour), int(minute), int(sec)
    )

    return date_time, int(nsec)


def _parse_date_full(date):
    """Parse full date string.

    This matches either 2014-12-12 17:29:43 or 2014-12-12T17:29:43.

    Args:
        date (str): the date string to be parsed.

    Returns:
        A tuple of the format (date_time, nsec), where date_time is a
        datetime.datetime object and nsec is 0.

    Raises:
        ValueError: if the date format does not match.
    """
    pattern = re.compile(
        r'^(?P<year>\d{4})-(?P<mon>[01]\d)-(?P<day>[0-3]\d)[\sTt]'
        r'(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})$'
    )

    if not pattern.match(date):
        raise ValueError('Wrong date format: {}'.format(date))

    year = pattern.search(date).group('year')
    month = pattern.search(date).group('mon')
    day = pattern.search(date).group('day')
    hour = pattern.search(date).group('hour')
    minute = pattern.search(date).group('min')
    sec = pattern.search(date).group('sec')
    nsec = 0

    date_time = datetime.datetime(
        int(year), int(month), int(day),
        int(hour), int(minute), int(sec)
    )

    return date_time, nsec


def _parse_date_time_with_nsec(date):
    """Parse time string with nanosecond resolution.

    This matches 17:29:43.802588035.

    Args:
        date (str): the date string to be parsed.

    Returns:
        A tuple of the format (date_time, nsec), where date_time is a
        datetime.time object and nsec is an int of the remaining
        nanoseconds.

    Raises:
        ValueError: if the date format does not match.
    """
    pattern = re.compile(
        r'^(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<nsec>\d{9})$'
    )

    if not pattern.match(date):
        raise ValueError('Wrong date format: {}'.format(date))

    hour = pattern.search(date).group('hour')
    minute = pattern.search(date).group('min')
    sec = pattern.search(date).group('sec')
    nsec = pattern.search(date).group('nsec')

    time = datetime.time(int(hour), int(minute), int(sec))

    return time, int(nsec)


def _parse_date_time(date):
    """Parse time string.

    This matches 17:29:43.

    Args:
        date (str): the date string to be parsed.

    Returns:
        A tuple of the format (date_time, nsec), where date_time is a
        datetime.time object and nsec is 0.

    Raises:
        ValueError: if the date format does not match.
    """
    pattern = re.compile(
        r'^(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})$'
    )

    if not pattern.match(date):
        raise ValueError('Wrong date format: {}'.format(date))

    hour = pattern.search(date).group('hour')
    minute = pattern.search(date).group('min')
    sec = pattern.search(date).group('sec')
    nsec = 0

    time = datetime.time(int(hour), int(minute), int(sec))

    return time, nsec


def _parse_date_timestamp(date):
    """Parse timestamp string in nanoseconds from epoch.

    This matches 1418423383802588035.

    Args:
        date (str): the date string to be parsed.

    Returns:
        A tuple of the format (date_time, nsec), where date_time is a
        datetime.datetime object and nsec is an int of the remaining
        nanoseconds.

    Raises:
        ValueError: if the date format does not match.
    """
    pattern = re.compile(r'^\d+$')

    if not pattern.match(date):
        raise ValueError('Wrong date format: {}'.format(date))

    timestamp_ns = int(date)

    date_time = datetime.datetime.fromtimestamp(
        timestamp_ns // NSEC_PER_SEC
    )
    # Set the microseconds to 0 because values < 1 second are covered
    # by the nsec value.
    date_time = date_time.replace(microsecond=0)
    nsec = timestamp_ns % NSEC_PER_SEC

    return date_time, nsec


def parse_date(date):
    """Try to parse a date string from one of many formats.

    Args:
        date (str): the date string to be parsed.

    Returns:
        A tuple of the format (date_time, nsec), where date_time is
        one of either datetime.datetime or datetime.time, depending on
        whether the date string contains full date information or only
        the time of day. The latter case can still be useful when used
        in conjuction with a trace collection's date to provide the
        missing information. The nsec element of the tuple is an int and
        corresponds to the nanoseconds for the given date/timestamp.
        This is due to datetime objects only supporting a resolution
        down to the microsecond.

    Raises:
        ValueError: if the date does not correspond to any of the
        supported formats.
    """
    parsers = [
        _parse_date_full_with_nsec, _parse_date_full,
        _parse_date_time_with_nsec, _parse_date_time,
        _parse_date_timestamp
    ]

    date_time = None
    nsec = None

    for parser in parsers:
        try:
            (date_time, nsec) = parser(date)
        except ValueError:
            continue

        # If no exception was raised, the parser found a match, so
        # stop iterating
        break

    if date_time is None or nsec is None:
        # None of the parsers were a match
        raise ValueError('Unrecognised date format: {}'.format(date))

    return date_time, nsec


def parse_trace_collection_date(collection, date, gmt=False, handles=None):
    """Parse a date string, using a trace collection to disambiguate
    incomplete dates.

    Args:
        collection (TraceCollection): a babeltrace TraceCollection
        instance.

        handles (TraceHandle): a babeltrace TraceHandle instance.

        date (string): the date string to be parsed.

        gmt (bool, optional): flag indicating whether the timestamp is
        in the local timezone or gmt (default: False).

    Returns:
        A timestamp (int) in nanoseconds since epoch, corresponding to
        the parsed date.

    Raises:
        ValueError: if the date format is unrecognised, or if the date
        format does not specify the date and the trace collection spans
        multiple days.
    """
    try:
        date_time, nsec = parse_date(date)
    except ValueError:
        # This might raise ValueError if the date is in an invalid
        # format, so just re-raise the exception to inform the caller
        # of the problem.
        raise

    # date_time will either be an actual datetime.datetime object, or
    # just a datetime.time object, depending on the format. In the
    # latter case, try and fill out the missing date information from
    # the trace collection's date.
    if isinstance(date_time, datetime.time):
        try:
            collection_date = trace_utils.get_trace_collection_date(collection,
                                                                    handles)
        except ValueError:
            raise ValueError(
                'Invalid date format for multi-day trace: {}'.format(date)
            )

        date_time = datetime.datetime.combine(collection_date, date_time)

    if gmt:
        date_time = date_time + datetime.timedelta(seconds=timezone)

    timestamp_ns = int(date_time.timestamp()) * NSEC_PER_SEC + nsec

    return timestamp_ns


def parse_trace_collection_time_range(collection, time_range,
                                      gmt=False, handles=None):
    """Parse a time range string, using a trace collection to
    disambiguate incomplete dates.

    Args:
        collection (TraceCollection): a babeltrace TraceCollection
        instance.

        handles (TraceHandle): a babeltrace TraceHandle instance.

        time_range (string): the time range string to be parsed.

        gmt (bool, optional): flag indicating whether the timestamps are
        in the local timezone or gmt (default: False).

    Returns:
        A tuple (begin, end) of the two timestamps (int) in nanoseconds
        since epoch, corresponding to the parsed dates.

    Raises:
        ValueError: if the time range or date format is unrecognised,
        or if the date format does not specify the date and the trace
        collection spans multiple days.
    """
    pattern = re.compile(r'^\[(?P<begin>.*),(?P<end>.*)\]$')
    if not pattern.match(time_range):
        raise ValueError('Invalid time range format: {}'.format(time_range))

    begin_str = pattern.search(time_range).group('begin').strip()
    end_str = pattern.search(time_range).group('end').strip()

    try:
        begin = parse_trace_collection_date(collection, begin_str,
                                            gmt, handles)
        end = parse_trace_collection_date(collection, end_str, gmt, handles)
    except ValueError:
        # Either of the dates was in the wrong format, propagate the
        # exception to the caller.
        raise

    return begin, end



# The MIT License (MIT)
#
# Copyright (C) 2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import namedtuple


PrioEvent = namedtuple('PrioEvent', ['timestamp', 'prio'])


class Stats():
    def reset(self):
        raise NotImplementedError()


class Process(Stats):
    def __init__(self, pid, tid, comm):
        self.pid = pid
        self.tid = tid
        self.comm = comm
        self.prio_list = []

    @classmethod
    def new_from_process(cls, proc):
        return cls(proc.pid, proc.tid, proc.comm)

    def update_prio(self, timestamp, prio):
        self.prio_list.append(PrioEvent(timestamp, prio))

    def reset(self):
        if self.prio_list:
            # Keep the last prio as the first for the next period
            self.prio_list = self.prio_list[-1:]


class IO(Stats):
    def __init__(self):
        # Number of bytes read or written
        self.read = 0
        self.write = 0

    def reset(self):
        self.read = 0
        self.write = 0

    def __iadd__(self, other):
        self.read += other.read
        self.write += other.write
        return self



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import sp, sv


class NetStateProvider(sp.StateProvider):
    def __init__(self, state):
        cbs = {
            'net_dev_xmit': self._process_net_dev_xmit,
            'netif_receive_skb': self._process_netif_receive_skb,
        }

        super().__init__(state, cbs)

    def _process_net_dev_xmit(self, event):
        self._state.send_notification_cb('net_dev_xmit',
                                         iface_name=event['name'],
                                         sent_bytes=event['len'],
                                         cpu_id=event['cpu_id'])

        cpu_id = event['cpu_id']
        if cpu_id not in self._state.cpus:
            return

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return

        proc = self._state.tids[cpu.current_tid]
        current_syscall = proc.current_syscall
        if current_syscall is None:
            return

        if proc.pid is not None and proc.pid != proc.tid:
            proc = self._state.tids[proc.pid]

        if current_syscall.name in sv.SyscallConsts.WRITE_SYSCALLS:
            # TODO: find a way to set fd_type on the write rq to allow
            # setting FD Type if FD hasn't yet been created
            fd = current_syscall.io_rq.fd
            if fd in proc.fds and proc.fds[fd].fd_type == sv.FDType.unknown:
                proc.fds[fd].fd_type = sv.FDType.maybe_net

    def _process_netif_receive_skb(self, event):
        self._state.send_notification_cb('netif_receive_skb',
                                         iface_name=event['name'],
                                         recv_bytes=event['len'],
                                         cpu_id=event['cpu_id'])



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import namedtuple


GraphDatum = namedtuple('GraphDatum', ['value', 'value_str'])
BarGraphDatum = namedtuple('BarGraphDatum', ['value', 'value_str', 'label'])
FreqGraphDatum = namedtuple(
    'FreqGraphDatum', ['value', 'value_str', 'lower_bound']
)


class Graph():
    MAX_GRAPH_WIDTH = 80
    BAR_CHAR = '█'
    HR_CHAR = '#'

    def __init__(self, data, get_value, get_value_str, title, unit):
        self._data = data
        self._get_value = get_value
        self._title = title
        self._unit = unit
        self._max_value = 0
        self._max_value_len = 0

        if get_value_str is not None:
            self._get_value_str_cb = get_value_str
        else:
            self._get_value_str_cb = Graph._get_value_str_default

    def _transform_data(self, data):
        graph_data = []

        for datum in data:
            graph_datum = self._get_graph_datum(datum)

            if graph_datum.value > self._max_value:
                self._max_value = graph_datum.value
            if len(graph_datum.value_str) > self._max_value_len:
                self._max_value_len = len(graph_datum.value_str)

            graph_data.append(graph_datum)

        return graph_data

    def _get_value_str(self, value):
        return self._get_value_str_cb(value)

    def _get_graph_datum(self, datum):
        value = self._get_value(datum)
        value_str = self._get_value_str(value)

        return GraphDatum(value, value_str)

    def _print_header(self):
        if self._title:
            print(self._title)

    def _print_separator(self):
        print(self.HR_CHAR * self.MAX_GRAPH_WIDTH)

    def _print_body(self):
        raise NotImplementedError()

    def print_graph(self):
        if not self._data:
            return

        self._print_header()
        self._print_separator()
        self._print_body()
        print()

    @staticmethod
    def _get_value_str_default(value):
        if isinstance(value, float):
            value_str = '{:0.02f}'.format(value)
        else:
            value_str = str(value)

        return value_str


class BarGraph(Graph):
    def __init__(self, data, get_value, get_label, get_value_str=None,
                 title=None, label_header=None, unit=None):
        super().__init__(data, get_value, get_value_str, title, unit)

        self._get_label = get_label
        self._label_header = label_header
        self._data = self._transform_data(self._data)

    def _get_graph_datum(self, datum):
        value = self._get_value(datum)
        value_str = self._get_value_str(value)
        label = self._get_label(datum)

        return BarGraphDatum(value, value_str, label)

    def _get_value_str(self, value):
        value_str = super()._get_value_str(value)
        if self._unit:
            value_str += ' ' + self._unit

        return value_str

    def _get_graph_header(self):
        if not self._label_header:
            return self._title

        title_len = len(self._title)
        space_width = (
            self.MAX_GRAPH_WIDTH - title_len + 1 + self._max_value_len + 1
        )

        return self._title + ' ' * space_width + self._label_header

    def _print_header(self):
        header = self._get_graph_header()
        print(header)

    def _get_bar_str(self, datum):
        if self._max_value == 0:
            bar_width = 0
        else:
            bar_width = int(self.MAX_GRAPH_WIDTH * datum.value /
                            self._max_value)
        space_width = self.MAX_GRAPH_WIDTH - bar_width
        bar_str = self.BAR_CHAR * bar_width + ' ' * space_width

        return bar_str

    def _print_body(self):
        for datum in self._data:
            bar_str = self._get_bar_str(datum)
            value_padding = ' ' * (self._max_value_len - len(datum.value_str))
            print(bar_str, value_padding + datum.value_str, datum.label)


class FreqGraph(Graph):
    LOWER_BOUND_WIDTH = 8

    def __init__(self, data, get_value, get_lower_bound,
                 get_value_str=None, title=None, unit=None):
        super().__init__(data, get_value, get_value_str, title, unit)

        self._get_lower_bound = get_lower_bound
        self._data = self._transform_data(self._data)

    def _get_graph_datum(self, datum):
        value = self._get_value(datum)
        value_str = self._get_value_str(value)
        lower_bound = self._get_lower_bound(datum)

        return FreqGraphDatum(value, value_str, lower_bound)

    def _print_header(self):
        header = self._title
        if self._unit:
            header += ' ({})'.format(self._unit)

        print(header)

    def _get_bar_str(self, datum):
        max_width = self.MAX_GRAPH_WIDTH - self.LOWER_BOUND_WIDTH
        if self._max_value == 0:
            bar_width = 0
        else:
            bar_width = int(max_width * datum.value / self._max_value)
        space_width = max_width - bar_width
        bar_str = self.BAR_CHAR * bar_width + ' ' * space_width

        return bar_str

    def _print_body(self):
        for datum in self._data:
            bound_str = FreqGraph._get_bound_str(datum)
            bar_str = self._get_bar_str(datum)
            value_padding = ' ' * (self._max_value_len - len(datum.value_str))
            print(bound_str, bar_str, value_padding + datum.value_str)

    @staticmethod
    def _get_bound_str(datum):
        return '{:>7.03f}'.format(datum.lower_bound)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#                      Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .analysis_test import AnalysisTest


class CpuTest(AnalysisTest):
    def write_trace(self):
        # runs the whole time: 100%
        self.trace_writer.write_sched_switch(1000, 5, 'swapper/5',
                                             0, 'prog100pc-cpu5', 42)
        # runs for 2s alternating with swapper out every 100ms
        self.trace_writer.sched_switch_50pc(1100, 5000, 0, 100, 'swapper/0',
                                            0, 'prog20pc-cpu0', 30664)
        # runs for 2.5s alternating with swapper out every 100ms
        self.trace_writer.sched_switch_50pc(5100, 10000, 1, 100, 'swapper/1',
                                            0, 'prog25pc-cpu1', 30665)
        # switch out prog100pc-cpu5
        self.trace_writer.write_sched_switch(11000, 5, 'prog100pc-cpu5',
                                             42, 'swapper/5', 0)
        self.trace_writer.flush()

    def test_cputop(self):
        test_name = 'cputop'
        expected = self.get_expected_output(test_name)
        result = self.get_cmd_output('lttng-cputop',
                                     options='--no-intersection')

        self._assertMultiLineEqual(result, expected, test_name)
